from .llamaindex_poma import (
    PomaFileReader,
    PomaChunksetNodeParser,
    PomaCheatsheetRetrieverLI,
)

__all__ = [
    "PomaFileReader",
    "PomaChunksetNodeParser",
    "PomaCheatsheetRetrieverLI",
]



from .langchain_poma import (
    PomaFileLoader,
    PomaChunksetSplitter,
    PomaCheatsheetRetrieverLC,
)


__all__ = [
    "PomaFileLoader",
    "PomaChunksetSplitter",
    "PomaCheatsheetRetrieverLC",
]



from .qdrant_poma import (
    PomaQdrant,
    QdrantConfig,
    VectorConfig,
    InferenceConfig,
    QdrantResponseError,
    SearchResult,
    chunk_uuid_string,
    DenseEmbedSync,
    SparseEmbedSync,
)

__all__ = [
    "PomaQdrant",
    "QdrantConfig",
    "VectorConfig",
    "InferenceConfig",
    "QdrantResponseError",
    "SearchResult",
    "chunk_uuid_string",
    "DenseEmbedSync",
    "SparseEmbedSync",
]



# retrieval.py
from collections import defaultdict
from itertools import chain
from typing import Any
import warnings


def deprecated(replacement: str):
    def decorator(func):
        msg = (
            f"{func.__name__}() is deprecated and will be removed in a future version. "
            f"Use {replacement} instead."
        )

        def wrapper(*args, **kwargs):
            warnings.warn(msg, DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        wrapper.__name__ = func.__name__
        wrapper.__doc__ = (func.__doc__ or "") + f"\n\nDEPRECATED: {msg}\n"
        return wrapper

    return decorator


def generate_cheatsheets(
    relevant_chunksets: list[dict[str, Any]],
    all_chunks: list[dict[str, Any]],
) -> list[dict[str, Any]]:
    # get chunks grouped by document file_id
    doc_chunks = defaultdict(list)
    for chunk in all_chunks:
        file_id = chunk.get("file_id") or chunk.get("tag") or "single_doc"
        chunk["file_id"] = file_id  # update
        doc_chunks[file_id].append(chunk)

    # Check for duplicate chunk_index values
    # (necessary when file_id was not set in chunks)
    for file_id, chunks in doc_chunks.items():
        chunk_indices = [c["chunk_index"] for c in chunks]
        if len(chunk_indices) != len(set(chunk_indices)):
            raise ValueError(f"Duplicate chunk_index found for file_id: {file_id}")

    # get relevant chunksets grouped by document file_id
    relevant_chunksets_per_doc = defaultdict(list)
    for chunkset in relevant_chunksets:
        file_id = chunkset.get("file_id") or chunkset.get("tag") or "single_doc"
        chunkset["file_id"] = file_id  # update
        if "chunks" not in chunkset:
            raise ValueError(
                "Chunkset not valid; must contain a 'chunks' key with a list of chunk IDs."
            )
        relevant_chunksets_per_doc[file_id].append(chunkset)

    # Ensure that chunksets and chunks correspond to the same file_ids
    for file_id in relevant_chunksets_per_doc.keys():
        if file_id not in doc_chunks:
            raise ValueError(
                f"Chunksets contain file_id '{file_id}' which is not present in the chunks."
            )

    # retrieve relevant chunks with content per document
    relevant_content_chunks: list[RetrievalChunk] = []
    for file_id, chunksets_per_doc in relevant_chunksets_per_doc.items():
        chunk_ids = list(  # flattened list
            chain.from_iterable(chunkset["chunks"] for chunkset in chunksets_per_doc)
        )
        relevant_chunks_dict = _get_relevant_chunks_for_ids(
            chunk_ids, doc_chunks[file_id]
        )
        relevant_chunks: list[RetrievalChunk] = chunks_from_dicts(relevant_chunks_dict)
        relevant_content_chunks.extend(relevant_chunks)

    return _cheatsheets_from_chunks(relevant_content_chunks)


@deprecated("generate_cheatsheets(relevant_chunksets, all_chunks)")
def generate_single_cheatsheet(
    relevant_chunksets: list[dict[str, Any]],
    all_chunks: list[dict[str, Any]],
) -> str:
    cheatsheets = generate_cheatsheets(
        relevant_chunksets=relevant_chunksets,
        all_chunks=all_chunks,
    )
    return cheatsheets[0].get("content", "") if cheatsheets else ""


########################
# RetrievalChunk Class #
########################


class RetrievalChunk:
    """
    Represents a chunk of text with associated metadata.
    Attributes:
        index (int): The index of the chunk within a sequence.
        file_id (str): The id associating the chunk with a document.
        content (str): The textual content of the chunk.
        depth_rebased (int, optional): The hierarchical depth of the chunk content.
            In cheatsheets, this affects indentation for certain text parts.
            Currently only used for code blocks.
    """

    def __init__(
        self,
        index: int,
        file_id: str,
        content: str,
        depth_rebased: int | None,
    ):
        self.index = index
        self.file_id = file_id
        self.content = content
        self.depth_rebased = depth_rebased

    @classmethod
    def from_chunk_dict(
        cls,
        chunk_dict: dict,
        block_min_depth: int | None,
    ):
        if block_min_depth is not None:
            depth = int(chunk_dict["depth"])
            depth_rebased = cls._rebase_depth(depth, block_min_depth)
        else:
            depth_rebased = None
        return cls(
            index=int(chunk_dict["chunk_index"]),
            file_id=str(
                chunk_dict.get("file_id") or chunk_dict.get("tag") or "single_doc"
            ),
            content=str(chunk_dict["content"]),
            depth_rebased=depth_rebased,
        )

    @staticmethod
    def _rebase_depth(depth: int, min_depth: int, base_unit: int = 0) -> int | None:
        rebased = depth - min_depth + base_unit
        return max(0, rebased)

    def __repr__(self):
        return f"RetrievalChunk(index={self.index}, file_id={self.file_id}, content={self.content}), depth_rebased={self.depth_rebased}"


def chunks_from_dicts(chunk_dicts: list[dict]) -> list[RetrievalChunk]:
    """
    Converts a list of chunk dictionaries into a list of Chunk objects.
    File_ids are needed to identify chunks from different documents;
    if is_single_doc is True, all chunks are assumed to come from a single document
    and file_id is optional.
    Args:
        chunk_dicts (list[dict]): A list of dictionaries, each representing a chunk with required keys:
            - "chunk_index": The index of the chunk within the document.
            - "file_id": The identifier of the document.
            - "content": The textual content of the chunk.
            - "depth": The depth or level of the chunk.
    Returns:
        list[Chunk]: A list of Chunk objects with the textual content needed for the cheatsheets.
    """

    # Determine the minimum depth per code block
    min_depth_per_code_block: dict[str, int] = {}
    for chunk_dict in chunk_dicts:
        block_id = chunk_dict.get("code")
        if block_id is None:
            continue
        depth = int(chunk_dict["depth"])
        current = min_depth_per_code_block.get(block_id)
        min_depth_per_code_block[block_id] = (
            depth if current is None else min(current, depth)
        )

    # Create Chunk objects
    all_chunks: list[RetrievalChunk] = []
    for chunk_dict in chunk_dicts:
        code_id = chunk_dict.get("code")
        if bool(code_id):
            block_min_depth = min_depth_per_code_block.get(str(code_id))
        else:
            block_min_depth = None
        chunk = RetrievalChunk.from_chunk_dict(chunk_dict, block_min_depth)
        all_chunks.append(chunk)

    # Sanity check: Make sure there are no duplicate chunk_index values
    check_dict = defaultdict(set)
    has_duplicates = any(
        chunk.index in check_dict[chunk.file_id]
        or check_dict[chunk.file_id].add(chunk.index)
        for chunk in all_chunks
    )
    if has_duplicates:
        raise ValueError(
            "Duplicate chunk indices found in single document mode. "
            "Each chunk must have a unique index."
        )

    return all_chunks


###################
# Private Methods #
###################


def _get_relevant_chunks_for_ids(
    chunk_ids: list[int],
    chunks: list[dict[str, Any]],
) -> list[dict[str, Any]]:
    chunk_indices_of_retrieved_chunksets = chunk_ids
    all_chunks_of_doc = chunks

    # Build helpers
    sorted_chunks = sorted(all_chunks_of_doc, key=lambda c: c["chunk_index"])
    index_to_chunk = {c["chunk_index"]: c for c in sorted_chunks}
    index_to_depth = {c["chunk_index"]: c["depth"] for c in sorted_chunks}

    # Find relatively deepest indices in the retrieval
    candidate_indices = set(chunk_indices_of_retrieved_chunksets)

    def is_ancestor(idx1, idx2):
        """True if idx1 is an ancestor of idx2."""
        # idx1 must be before idx2 and have smaller depth
        if idx1 >= idx2:
            return False
        if idx1 not in index_to_depth or idx2 not in index_to_depth:
            return False
        depth1 = index_to_depth[idx1]
        depth2 = index_to_depth[idx2]
        if depth1 >= depth2:
            return False
        # scan from idx1+1 up to idx2, making sure all are deeper than depth1 until idx2
        # Use sorted_chunks to iterate in order, but check chunk_index values
        for chunk in sorted_chunks:
            chunk_idx = chunk["chunk_index"]
            if chunk_idx <= idx1:
                continue
            if chunk_idx > idx2:
                break
            if chunk_idx == idx2:
                continue  # skip idx2 itself, we'll check it after
            depth = index_to_depth[chunk_idx]
            if depth <= depth1:
                return False
        return True

    # Exclude any index that is an ancestor of another in the set
    relatively_deepest = set(candidate_indices)
    for idx1 in candidate_indices:
        for idx2 in candidate_indices:
            if idx1 != idx2 and is_ancestor(idx1, idx2):
                relatively_deepest.discard(idx1)
                break

    # Standard subtree/parent finding routines
    def get_child_indices(chunk_index: int) -> list[int]:
        if chunk_index not in index_to_depth:
            return []
        base_depth = index_to_depth[chunk_index]
        children = []
        # Find the position of chunk_index in sorted_chunks
        start_pos = None
        for pos, chunk in enumerate(sorted_chunks):
            if chunk["chunk_index"] == chunk_index:
                start_pos = pos
                break
        if start_pos is None:
            return []
        # Iterate forward from the next position
        for i in range(start_pos + 1, len(sorted_chunks)):
            idx = sorted_chunks[i]["chunk_index"]
            depth = sorted_chunks[i]["depth"]
            if depth <= base_depth:
                break
            children.append(idx)
        return children

    def get_parent_indices(chunk_index: int) -> list[int]:
        if chunk_index not in index_to_depth:
            return []
        parents = []
        current_depth = index_to_depth[chunk_index]
        # Find the position of chunk_index in sorted_chunks
        start_pos = None
        for pos, chunk in enumerate(sorted_chunks):
            if chunk["chunk_index"] == chunk_index:
                start_pos = pos
                break
        if start_pos is None:
            return []
        # Iterate backward from the previous position
        for i in range(start_pos - 1, -1, -1):
            idx = sorted_chunks[i]["chunk_index"]
            depth = sorted_chunks[i]["depth"]
            if depth < current_depth:
                parents.append(idx)
                current_depth = depth
        return parents[::-1]  # root -> leaf order

    # Collect all relevant indices
    all_indices = set(
        chunk_indices_of_retrieved_chunksets
    )  # always include all search hits
    for idx in relatively_deepest:
        all_indices.update(get_child_indices(idx))

    # Parents for all found nodes
    for idx in list(all_indices):
        all_indices.update(get_parent_indices(idx))

    # Return in doc order
    return [index_to_chunk[i] for i in sorted(all_indices)]


def _cheatsheets_from_chunks(
    content_chunks: list[RetrievalChunk],
) -> list[dict[str, Any]]:
    if not content_chunks:
        return []

    if isinstance(content_chunks[0], dict):
        raise ValueError(
            "Input to _cheatsheets_from_chunks must be a list of RetrievalChunk objects, not dicts."
            "Use chunks_from_dicts() to convert dicts to RetrievalChunk objects first."
        )

    def _format_chunk_content(chunk: "RetrievalChunk") -> str:
        if not getattr(chunk, "depth_rebased", False):
            return chunk.content
        else:
            indent = " " * 4 * (chunk.depth_rebased or 0)
            return f"{indent}{chunk.content}"

    cheatsheets: list[dict] = []

    compressed_data = {}
    content_chunks = sorted(content_chunks, key=lambda c: (c.file_id, c.index))
    for chunk in content_chunks:
        if chunk.file_id not in compressed_data:
            # If there is data stored for a previous file_id, save it to the cheatsheets list
            if compressed_data:
                for key, value in compressed_data.items():
                    cheatsheets.append(
                        {"file_id": key, "tag": key, "content": value["content"]}
                    )
            # Clear the compressed_data for the current file_id
            compressed_data.clear()
            # Start a new entry for the current file_id
            compressed_data[chunk.file_id] = {
                "content": _format_chunk_content(chunk),
                "last_chunk": chunk.index,
            }
        else:
            chunk_content = _format_chunk_content(chunk)
            # Check if chunks are consecutive
            if chunk.index == int(compressed_data[chunk.file_id]["last_chunk"]) + 1:
                compressed_data[chunk.file_id]["content"] += "\n" + chunk_content
            else:
                compressed_data[chunk.file_id]["content"] += "\n[…]\n" + chunk_content
            # Update the last chunk index
            compressed_data[chunk.file_id]["last_chunk"] = chunk.index

    # Save the last processed entry to the cheatsheets list
    if compressed_data:
        for key, value in compressed_data.items():
            cheatsheets.append(
                {"file_id": key, "tag": key, "content": value["content"]}
            )

    return cheatsheets



# client.py
import os
import httpx
import zipfile
import io
import json
import time
from pathlib import Path
from typing import Any

from poma.exceptions import (
    AuthenticationError,
    RemoteServerError,
    InvalidResponseError,
)
from poma.retrieval import generate_cheatsheets


USER_AGENT = "poma-ai-sdk/0.1.0"

API_BASE_URL = "https://api.poma-ai.com/api/v1"


def extract_chunks_and_chunksets_from_poma_archive(
    poma_archive_data: bytes | None = None,
    poma_archive_path: str | os.PathLike[str] | None = None,
) -> dict[str, Any]:
    """
    Extract chunks and chunksets from a POMA archive file.
    POMA archive is a zip file containing chunks.json and chunksets.json.
    Args:
        poma_archive_data: The POMA archive as bytes.
        poma_archive_path: Path to the POMA archive file.
    Returns:
        dict: A dictionary with ``chunks`` and ``chunksets`` keys.
    """
    chunks = None
    chunksets = None
    if poma_archive_path:
        with zipfile.ZipFile(poma_archive_path, "r") as zip_ref:
            chunks = zip_ref.read("chunks.json")
            chunksets = zip_ref.read("chunksets.json")
    elif poma_archive_data:
        with zipfile.ZipFile(io.BytesIO(poma_archive_data), "r") as zip_ref:
            chunks = zip_ref.read("chunks.json")
            chunksets = zip_ref.read("chunksets.json")
    else:
        raise ValueError(
            "Either poma_archive_data or poma_archive_path must be provided."
        )
    if not chunks or not chunksets:
        raise KeyError("Result must contain 'chunks' and 'chunksets' keys.")
    return {"chunks": json.loads(chunks), "chunksets": json.loads(chunksets)}


class Poma:
    """
    Client for interacting with the POMA API.
    """

    def __init__(
        self,
        api_key: str | None = None,
        *,
        timeout: float = 600.0,
        client: httpx.Client | None = None,
    ):
        """
        Initialize the POMA client.
        Args:
            api_key (str, optional):
                API key for authenticating with POMA. If not provided,
                the value is read from the environment variable `POMA_API_KEY`.
            timeout (float, default=600.0):
                Timeout (in seconds) for all HTTP requests.
            client (httpx.Client, optional):
                Custom HTTP client. If not provided, a default client is created.
        """
        api_base_url = API_BASE_URL
        # Override API base URL if environment variable is set
        if os.environ.get("API_BASE_URL"):
            api_base_url = os.environ.get("API_BASE_URL")
        if not api_base_url:
            raise ValueError("API base URL cannot be empty.")

        self.base_api_url = api_base_url.rstrip("/")
        self._client = client or httpx.Client(
            timeout=timeout, headers={"user-agent": USER_AGENT}
        )
        if not (api_key := api_key or os.environ.get("POMA_API_KEY", "")):
            raise Exception("POMA_API_KEY environment variable not set.")
        self._client.headers.update({"Authorization": f"Bearer {api_key}"})

    def start_chunk_file(
        self,
        file_path: str | os.PathLike[str],
        *,
        base_url: str | None = None,
    ) -> dict[str, Any]:
        """
        Submit a file with text to POMA for chunking.
        Args:
            file_path (str | os.PathLike[str]):
                Path to the input file (string or path-like). Must have an allowed file extension.
            base_url (str, optional):
                Optional base URL to resolve relative links within the file.
        Returns:
            A dictionary containing a unique job identifier for the submitted job.
        """
        if file_path is None or (isinstance(file_path, str) and not file_path.strip()):
            raise ValueError("file_path must be a non-empty string or path-like.")
        path = Path(file_path)
        payload = {}
        payload["is_sdk"] = True
        if base_url:
            payload["base_url"] = base_url
        try:
            response = self._client.post(
                f"{self.base_api_url}/ingest",
                data=payload,
                files={
                    "file": (path.name, path.read_bytes()),
                },
            )
            response.raise_for_status()
        except httpx.HTTPStatusError as error:
            status = error.response.status_code
            if status in (401, 403):
                raise AuthenticationError(
                    f"Failed to submit file '{path}': authentication error"
                ) from error
            raise RemoteServerError(
                f"Failed to submit file '{path}': {status}"
            ) from error
        try:
            data = response.json()
        except ValueError as error:
            raise InvalidResponseError(
                "Server returned non-JSON or empty body"
            ) from error
        return data

    def get_chunk_result(
        self,
        job_id: str,
        *,
        initial_delay: float = 5.0,
        poll_interval: float = 3.0,
        max_interval: float = 15.0,
        show_progress: bool = False,
        download_dir: str | os.PathLike[str] | None = None,
        filename: str | None = None,
    ) -> dict[str, Any]:
        """
        Poll POMA for the result of a chunking job until completion.
        Args:
            job_id (str):
                The unique identifier of the submitted job.
            initial_delay (float, default=5.0):
                Initial delay (in seconds) before the first poll request.
            poll_interval (float, default=1.0):
                Starting interval (in seconds) between polling requests.
            max_interval (float, default=15.0):
                Maximum interval (in seconds) between polling requests.
            show_progress (bool, default=False):
                If True, logs progress messages during polling.
            download_dir (str | os.PathLike[str], optional):
                Directory to save the downloaded file in. If neither download_dir nor
                filename is set, the result is returned in memory (no file saved). If
                filename is set but download_dir is not, the file is saved in the
                current directory.
            filename (str, optional):
                Name for the saved .poma file. If it does not end with ``.poma``, that
                suffix is appended. If not set when saving to disk, uses the server
                filename when provided, otherwise ``{job_id}.poma``.
        Returns:
            The JSON result containing at least the keys `chunks` and `chunksets`.

        """
        time.sleep(initial_delay)
        current_interval = poll_interval
        last_status = None

        while True:
            time.sleep(current_interval)
            try:
                response = self._client.get(f"{self.base_api_url}/jobs/{job_id}/status")
                response.raise_for_status()
                data = response.json()
                status = data.get("status", "")
                if status == "done":
                    download = data.get("download", {})
                    download_url = download.get("download_url", "")
                    if not download_url:
                        raise RuntimeError(
                            "Failed to receive download URL from server."
                        )

                    if download_dir is None and filename is None:
                        # Return bytes content instead of saving to file
                        file_bytes = self.download_bytes(download_url)
                        return self.extract_chunks_and_chunksets_from_poma_archive(
                            poma_archive_data=file_bytes
                        )
                    else:
                        # Save downloaded file (to download_dir or current dir if only filename set)
                        save_filename = (
                            filename or download.get("filename") or f"{job_id}.poma"
                        )
                        if not save_filename.endswith(".poma"):
                            save_filename = f"{save_filename}.poma"
                        save_dir = (
                            download_dir
                            if download_dir not in (None, "")
                            else "."
                        )
                        downloaded_file_path = self.download_file(
                            download_url, save_filename, save_directory=save_dir
                        )
                        return self.extract_chunks_and_chunksets_from_poma_archive(
                            poma_archive_path=downloaded_file_path
                        )
                elif status == "failed":
                    error_code = data.get("code", "unknown")
                    error_details = data.get("error", "No details provided.")
                    error_message = (
                        f"Job failed with code {error_code}: {error_details}"
                    )
                    raise RemoteServerError(
                        f"Job failed: {data.get('error', error_message)}"
                    )
                elif status == "processing":
                    if show_progress:
                        print(f"Job {job_id} is still processing...")
                    if last_status == "pending":
                        current_interval = poll_interval
                    current_interval = min(current_interval * 1.5, max_interval)
                elif status == "pending":
                    if show_progress:
                        print(
                            f"Job {job_id} is pending (queued due to rate limiting, sequential processing - common on demo accounts)..."
                        )
                    current_interval = min(current_interval * 1.5, max_interval)
                else:
                    raise InvalidResponseError(f"Unexpected job status: {status}")
                last_status = status
            except httpx.HTTPStatusError as error:
                raise RemoteServerError(
                    f"HTTP error: {error.response.status_code} {error.response.text}"
                ) from error
            except Exception as error:
                raise RuntimeError(f"POMA-AI job polling failed: {error}") from error

    def extract_chunks_and_chunksets_from_poma_archive(
        self,
        poma_archive_data: bytes | None = None,
        poma_archive_path: str | os.PathLike[str] | None = None,
    ) -> dict[str, Any]:
        """
        Extract chunks and chunksets from a POMA archive; delegates to module-level function.
        POMA archive is a zip file containing chunks.json and chunksets.json.
        Args:
            poma_archive_data: The POMA archive as bytes.
            poma_archive_path: Path to the POMA archive file.
        Returns:
            dict: A dictionary with ``chunks`` and ``chunksets`` keys.
        """
        return extract_chunks_and_chunksets_from_poma_archive(
            poma_archive_data=poma_archive_data,
            poma_archive_path=poma_archive_path,
        )

    def create_cheatsheet(
        self,
        relevant_chunksets: list[dict[str, Any]],
        all_chunks: list[dict[str, Any]],
    ) -> str:
        """
        Generates a single cheatsheet for one single document
        from relevant chunksets (relevant for a certain query)
        and from all chunks of that document (providing the textual content).
        Args:
            relevant_chunksets (list[dict]): A list of chunksets, each containing a "chunks" key with a list of chunk IDs.
            all_chunks (list[dict]): A list of all chunk dictionaries of the same document, each representing a chunk of content.
        Returns:
            str: The textual content of the generated cheatsheet.
        """
        cheatsheets = generate_cheatsheets(relevant_chunksets, all_chunks)
        if (
            not cheatsheets
            or not isinstance(cheatsheets, list)
            or len(cheatsheets) == 0
            or "content" not in cheatsheets[0]
        ):
            raise Exception(
                "Unknown error; cheatsheet could not be created from input chunks."
            )
        return cheatsheets[0]["content"]

    def create_cheatsheets(
        self,
        relevant_chunksets: list[dict[str, Any]],
        all_chunks: list[dict[str, Any]],
    ) -> list[dict[str, Any]]:
        """
        Generates cheatsheets from relevant chunksets (relevant for a certain query)
        and from all the chunks of all affected documents (providing the textual content).
        One cheatsheet is created for each document found in the chunks (tagged with file_id).
        Args:
            relevant_chunksets (list[dict]): A list of chunksets, each containing a "chunks" key with a list of chunk IDs.
            all_chunks (list[dict]): A list of all available chunk dictionaries of affected documents, each representing a chunk of content.
        Returns:
            list[dict]: A list of dictionaries representing the generated cheatsheets, each containing:
                - 'file_id': The tag associated with the respective document.
                - 'content': The textual content of the generated cheatsheet.
        """
        return generate_cheatsheets(relevant_chunksets, all_chunks)

    def download_file(
        self,
        download_url: str,
        filename: str | None = None,
        *,
        save_directory: str | os.PathLike[str] | None = None,
    ) -> str:
        """
        Download a file from the given download URL.
        Args:
            download_url (str):
                The URL to download the file from.
            filename (str, optional):
                The filename to save the file as. If not provided, will be extracted from URL.
            save_directory (str | os.PathLike[str], optional):
                Directory to save the file in. If not provided, saves to current directory.
        Returns:
            str: The path to the downloaded file.
        """
        if not download_url:
            raise ValueError("download_url cannot be empty")

        # Determine filename
        if not filename:
            filename = Path(download_url).name or "downloaded_file"

        # Determine save directory (default current directory so path has a parent)
        if save_directory:
            save_path = Path(save_directory) / filename
        else:
            save_path = Path(".") / filename

        # Create the directory if it doesn't exist (skip when file is in cwd)
        parent = os.path.dirname(save_path)
        if parent:
            os.makedirs(parent, exist_ok=True)

        # Download the file data
        content = self.download_bytes(download_url)

        # Save the file
        with open(save_path, "wb") as f:
            f.write(content)

        return str(save_path)

    def download_bytes(
        self,
        download_url: str,
    ) -> bytes:
        """
        Download a file from the given download URL and return the bytes content.
        Args:
            download_url (str):
                The URL to download the file from.
        Returns:
            bytes: The content of the downloaded file as bytes.
        """
        if not download_url:
            raise ValueError("download_url cannot be empty")

        # Construct the full URL if it's a relative path
        if download_url.startswith("/"):
            full_url = f"{self.base_api_url}{download_url}"
        else:
            full_url = download_url

        print("Downloading file from:", full_url)
        try:
            response = self._client.get(full_url)
            response.raise_for_status()
            return response.content
        except httpx.HTTPStatusError as error:
            status = error.response.status_code
            if status in (401, 403):
                raise AuthenticationError(
                    f"Failed to download '{download_url}': authentication error"
                ) from error
            raise RemoteServerError(
                f"Failed to download '{download_url}': {status}"
            ) from error
        except Exception as error:
            raise RuntimeError(f"File download failed: {error}") from error

    def close(self):
        self._client.close()

    def __enter__(self) -> "Poma":
        return self

    def __exit__(self, *exc):
        self.close()



# ---------------------------------------------------------------------
#  POMA integration for LangChain
# ---------------------------------------------------------------------

import os
import hashlib
from typing import Any
from pathlib import Path
from collections.abc import Iterable
from concurrent.futures import ThreadPoolExecutor, as_completed

from langchain.document_loaders.base import BaseLoader
from langchain.schema import Document
from langchain.schema.retriever import BaseRetriever
from langchain_core.vectorstores import VectorStore
from langchain_core.callbacks.manager import CallbackManagerForRetrieverRun
from langchain_text_splitters import TextSplitter
from pydantic import Field, PrivateAttr

from poma import Poma
from poma.exceptions import InvalidInputError
from poma.retrieval import chunks_from_dicts, _cheatsheets_from_chunks

__all__ = ["PomaFileLoader", "PomaChunksetSplitter", "PomaCheatsheetRetrieverLC"]


# ------------------------------------------------------------------ #
#  Load from Path → LC Documents                                     #
# ------------------------------------------------------------------ #


class PomaFileLoader(BaseLoader):

    def __init__(self, input_path: str | Path):
        """Initialize with a file or directory path."""
        self.input_path = Path(input_path).expanduser().resolve()

    def load(self) -> list[Document]:
        """
        Load files from the input path (file or directory) into LangChain Documents.
        Only files with allowed extensions are processed; others are skipped.
        """
        path = self.input_path
        if not path.exists():
            raise FileNotFoundError(f"No such path: {path}")

        documents: list[Document] = []
        skipped: int = 0

        def _process_file(file_path: Path):
            nonlocal skipped, documents
            if not file_path.is_file():
                return
            file_bytes = file_path.read_bytes()
            file_hash = hashlib.md5(file_bytes).hexdigest()
            if file_path.suffix.lower() == ".pdf":
                page_content: str = ""  # LangChain requires str
            else:
                try:
                    page_content = file_bytes.decode("utf-8")
                except UnicodeDecodeError:
                    skipped += 1
                    return
            documents.append(
                Document(
                    page_content=page_content,
                    metadata={
                        "source_path": str(file_path),
                        "doc_id": f"{file_hash}",
                    },
                )
            )

        if path.is_file():
            _process_file(path)
        elif path.is_dir():
            for path_in_dir in sorted(path.rglob("*")):
                _process_file(path_in_dir)
        else:
            raise FileNotFoundError(f"Unsupported path type (not file/dir): {path}")

        if not documents:
            raise InvalidInputError(f"No supported files found.")
        if skipped > 0:
            print(f"Skipped {skipped} file(s) due to unsupported or unreadable type.")
        return documents


# ------------------------------------------------------------------ #
#  Generate Chunksets                                                #
# ------------------------------------------------------------------ #


class PomaChunksetSplitter(TextSplitter):

    _client: Poma = PrivateAttr()
    _show_progress: bool = PrivateAttr(default=False)

    def __init__(self, client: Poma, *, verbose: bool = False, **kwargs):
        """Initialize with a Poma client and optional verbosity."""
        super().__init__(**kwargs)
        self._client = client
        self._show_progress = bool(verbose)

    def split_text(self, text: str) -> list[str]:
        """Not implemented, use split_documents()."""
        raise NotImplementedError("Not implemented, use split_documents().")

    def split_documents(self, documents: Iterable[Document]) -> list[Document]:
        """
        Split LangChain Documents into chunkset Documents via POMA API.
        Each output Document corresponds to a chunkset, with associated chunks in metadata.
        """
        documents = list(documents)
        if not documents:
            raise InvalidInputError("No documents provided to split.")

        total_docs = len(documents)
        chunked_docs: list[Document] = []
        failed_paths: list[str] = []

        def _safe_int(value: object) -> int | None:
            if isinstance(value, bool):
                return None
            if isinstance(value, int):
                return value
            if isinstance(value, str):
                try:
                    return int(value.strip())
                except Exception:
                    return None
            try:
                return int(value)  # type: ignore[arg-type]
            except Exception:
                return None

        def _doc_id_and_src(doc: Document) -> tuple[str, str]:
            src_path = doc.metadata.get("source_path", "in-memory-text")
            doc_id = doc.metadata.get("doc_id") or Path(src_path).stem or "unknown-doc"
            return doc_id, src_path

        def _process_one(
            poma_doc: Document, doc_idx: int
        ) -> tuple[list[Document], str | None]:
            """Process a single document via POMA API, return chunked Documents or failed source path."""
            try:
                doc_id, src_path = _doc_id_and_src(poma_doc)
                path_obj = None
                if src_path and src_path.strip() and isinstance(src_path, str):
                    try:
                        path = Path(src_path).resolve()
                        if path.exists():
                            path_obj = path
                    except Exception:
                        path_obj = None
                if not path_obj:
                    raise InvalidInputError(
                        "No valid source_path found in document metadata."
                    )
                start_result = self._client.start_chunk_file(path_obj, base_url=None)
                job_id = start_result.get("job_id")
                if not job_id:
                    raise RuntimeError("Failed to receive job ID from server.")
                if self._show_progress:
                    print(
                        f"[{doc_idx}/{total_docs}] ⏳ Job {job_id} started for: {src_path}. Polling for results..."
                    )
                result = self._client.get_chunk_result(
                    str(job_id), show_progress=self._show_progress
                )
                chunks: list[dict] = result.get("chunks", [])
                chunksets: list[dict] = result.get("chunksets", [])
            except Exception as exception:
                print(
                    f"[{doc_idx}/{total_docs}] ❌ Exception chunking document: {exception}"
                )
                src_path = poma_doc.metadata.get("source_path", "in-memory-text")
                return [], src_path

            file_docs: list[Document] = []
            try:
                chunks_by_index: dict[int, dict] = {}
                for chunk in chunks:
                    idx = _safe_int(chunk.get("chunk_index"))
                    if idx is not None:
                        chunks_by_index[idx] = chunk
                for cs in chunksets:
                    chunkset_index = cs.get("chunkset_index")
                    chunks_indices = cs.get("chunks", []) or []
                    normalized_indices: list[int] = []
                    for chunk_index in chunks_indices:
                        idx = _safe_int(chunk_index)
                        if idx is not None:
                            normalized_indices.append(idx)
                    relevant_chunks = [
                        chunks_by_index[idx]
                        for idx in normalized_indices
                        if idx in chunks_by_index
                    ]
                    file_docs.append(
                        Document(
                            page_content=cs.get("contents", ""),
                            metadata={
                                "doc_id": doc_id,
                                "chunkset_index": chunkset_index,
                                "chunkset": cs,
                                "chunks": relevant_chunks,
                                "source_path": src_path,
                            },
                        )
                    )
            except Exception as exception:
                print(
                    f"[{doc_idx}/{total_docs}] ❌ Exception processing chunking result: {exception}"
                )
                src_path = poma_doc.metadata.get("source_path", "in-memory-text")
                return [], src_path
            return file_docs, None

        # parallel processing of documents
        cores = os.cpu_count() or 1
        group_size = 5 if cores >= 5 else cores
        for start in range(0, total_docs, group_size):
            batch = list(
                enumerate(documents[start : start + group_size], start=start + 1)
            )
            with ThreadPoolExecutor(max_workers=group_size) as executor:
                futures = {
                    executor.submit(_process_one, doc, idx): (idx, doc)
                    for idx, doc in batch
                }
                for future in as_completed(futures):
                    idx, doc = futures[future]
                    try:
                        doc_as_chunk, failed_src = future.result()
                        if failed_src is None:
                            chunked_docs.extend(doc_as_chunk)
                            if self._show_progress:
                                src_path = doc.metadata.get(
                                    "source_path", "in-memory-text"
                                )
                                print(
                                    f"[{idx}/{total_docs}] ✅ Done: {src_path} (+{len(doc_as_chunk)} doc-chunks)"
                                )
                        else:
                            failed_paths.append(failed_src)
                            if self._show_progress:
                                print(f"[{idx}/{total_docs}] ❌ Failed: {failed_src}")
                    except Exception as error:
                        failed_paths.append(
                            doc.metadata.get("source_path", "in-memory-text")
                        )
                        if self._show_progress:
                            print(
                                f"[{idx}/{total_docs}] ❌ Failed with unexpected error: {error}"
                            )

        if failed_paths:
            print("The following files failed to process:")
            for path in failed_paths:
                print(f" - {path}")

        if not chunked_docs:
            raise InvalidInputError("No documents could be split successfully.")

        return chunked_docs


# ------------------------------------------------------------------ #
#  Cheatsheet retriever                                              #
# ------------------------------------------------------------------ #


class PomaCheatsheetRetrieverLC(BaseRetriever):

    tags: list[str] | None = Field(default=None)
    metadata: dict[str, Any] | None = Field(default=None)

    _vector_store: VectorStore = PrivateAttr()
    _top_k: int = PrivateAttr()

    def __init__(
        self,
        vector_store: VectorStore,
        *,
        top_k: int = 6,
        **kwargs,
    ):
        """Initialize with a VectorStore and number of top_k results to retrieve."""
        super().__init__(**kwargs)
        self._vector_store = vector_store
        self._top_k = top_k

    def _retrieve(self, query: str) -> list[Document]:
        """Retrieve chunkset documents and generate cheatsheets for the given query."""
        hits = self._vector_store.similarity_search(query, k=self._top_k)
        if not hits:
            return []
        grouped: dict[str, list[Document]] = {}
        for doc in hits:
            doc_id = doc.metadata["doc_id"]
            grouped.setdefault(doc_id, []).append(doc)
        cheatsheet_docs: list[Document] = []
        for doc_id, chunked_docs in grouped.items():
            cheatsheet = self._create_cheatsheet_langchain(chunked_docs)
            cheatsheet_docs.append(
                Document(page_content=cheatsheet, metadata={"doc_id": doc_id})
            )
        return cheatsheet_docs

    def _get_relevant_documents(
        self, query: str, *, run_manager: CallbackManagerForRetrieverRun
    ) -> list[Document]:
        """Retrieve relevant documents with callback management."""
        try:
            documents = self._retrieve(query)
            run_manager.on_retriever_end(documents)
            return documents
        except Exception as exception:
            run_manager.on_retriever_error(exception)
            raise

    def _create_cheatsheet_langchain(self, chunked_docs: list[Document]) -> str:
        """Generate a single deduplicated cheatsheet from chunked documents."""
        all_chunk_dicts = []
        seen = set()
        for doc in chunked_docs:
            doc_id = doc.metadata.get("doc_id", "unknown_doc")
            for chunk in doc.metadata.get("chunks", []):
                if not isinstance(chunk, dict):
                    continue
                chunk_index = chunk["chunk_index"]
                if chunk_index not in seen:
                    seen.add(chunk_index)
                    chunk["tag"] = doc_id
                    all_chunk_dicts.append(chunk)
        all_chunks = chunks_from_dicts(all_chunk_dicts)
        cheatsheets = _cheatsheets_from_chunks(all_chunks)
        if (
            not cheatsheets
            or not isinstance(cheatsheets, list)
            or len(cheatsheets) == 0
            or "content" not in cheatsheets[0]
        ):
            raise Exception(
                "Unknown error; cheatsheet could not be created from input chunks."
            )
        return cheatsheets[0]["content"]



# ---------------------------------------------------------------------
#  POMA integration for Qdrant
# ---------------------------------------------------------------------
"""
Sync POMA–Qdrant integration with dense, sparse, and hybrid search.

- BasePomaQdrantCore: shared logic (prepare points, RRF fusion, cheatsheet inputs).
- SyncClientOps: thin Qdrant I/O adapter.
- PomaQdrant: supports fastembed, external_*, and inference_* modes.
- fastembed mode is environment-dependent: it requires a Qdrant client with FastEmbed integration
  (upsert/query_points with models.Document). For production, pass a pre-configured client
  or use embedding_mode='external_dense' (or external_*).
"""

import inspect
import os
import uuid
from collections.abc import Callable, Sequence
from dataclasses import dataclass

from qdrant_client import QdrantClient, models as qcm
from qdrant_client.http.exceptions import UnexpectedResponse
from qdrant_client.http import models as qmodels

from poma.client import extract_chunks_and_chunksets_from_poma_archive
from poma.retrieval import generate_cheatsheets

DEFAULT_CLOUD_DENSE_MODEL = "openai/text-embedding-3-large"
DEFAULT_CLOUD_DENSE_DIMENSIONS = 1536
DEFAULT_CLOUD_SPARSE_MODEL = "Qdrant/bm25"


__all__ = [
    "PomaQdrant",
    "QdrantConfig",
    "VectorConfig",
    "InferenceConfig",
    "QdrantResponseError",
    "SearchResult",
    "chunk_uuid_string",
    "DenseEmbedSync",
    "SparseEmbedSync",
]


def _chunk_data_from_input(
    chunk_data: dict | str | os.PathLike[str],
) -> dict:
    """Accept chunk_data dict or path to a .poma file; return chunk_data dict."""
    if isinstance(chunk_data, (str, os.PathLike)):
        path = os.fspath(chunk_data)
        if not path.lower().endswith(".poma"):
            raise ValueError(
                "Path must point to a .poma file; got {!r}".format(path)
            )
        return extract_chunks_and_chunksets_from_poma_archive(
            poma_archive_path=chunk_data
        )
    return chunk_data

@dataclass
class SearchResult:
    """Type-safe search result; matches the dict shape returned by search()."""
    id: str
    score: float
    text: str
    metadata: dict


class QdrantResponseError(RuntimeError):
    """SDK-level wrapper for qdrant-client UnexpectedResponse errors."""

    def __init__(
        self,
        message: str,
        *,
        status_code: int | None = None,
        reason_phrase: str | None = None,
        raw_content: str | None = None,
    ) -> None:
        super().__init__(message)
        self.status_code = status_code
        self.reason_phrase = reason_phrase
        self.raw_content = raw_content


def chunk_uuid_string(file_id: str, chunkset_index: int) -> str:
    """
    Deterministic RFC4122 UUID (v5) from (file_id, chunkset_index).
    Same inputs -> same UUID. Different index -> different UUID.
    """
    if chunkset_index < 0:
        raise ValueError("chunkset_index must be non-negative")
    name = f"{file_id}#{chunkset_index}"
    return str(uuid.uuid5(uuid.NAMESPACE_URL, name))


# Dense: one float-vector per text.
# Callables may accept Sequence[str] for flexibility.
DenseEmbedSync = Callable[[Sequence[str]], list[list[float]]]

# Sparse: one sparse vector per text.
# Accepted per-item return types:
#   - qmodels.SparseVector(indices=[...], values=[...])
#   - tuple[list[int], list[float]]  -> (indices, values)
#   - dict with keys {"indices": [...], "values": [...]}
SparseEmbedSync = Callable[[Sequence[str]], list[object]]


@dataclass(frozen=True)
class QdrantConfig:
    """
    Qdrant deployment options:
      - mode="memory": in-process, RAM-only (sync only)
      - mode="path":   in-process, persistent folder on disk (sync only)
      - mode="url":    HTTP endpoint (local server or cloud) (sync)
      - api_key:      API key for cloud inference (required if cloud_inference is True)
      - cloud_inference: if True, use cloud inference (default is False)
      - client_kwargs: optional pass-through kwargs for QdrantClient(...)
        (e.g. {"timeout": 120, "pool_size": 20})
    """
    mode: str = "memory"          # "memory" | "path" | "url"
    path: str | None = None       # required if mode="path"
    url: str | None = None        # required if mode="url"
    api_key: str | None = None    # optional (cloud)
    cloud_inference: bool = False  # optional (cloud)
    client_kwargs: dict[str, object] | None = None


@dataclass(frozen=True)
class VectorConfig:
    """
    Vector/index configuration used when embedding_mode is external_* or inference_* (we create collections explicitly).
    """
    distance: qmodels.Distance = qmodels.Distance.COSINE
    dense_size: int | None = None          # required for external_dense/external_hybrid and inference_dense/inference_hybrid
    sparse_name: str = "sparse"            # name for sparse vector space
    dense_name: str = "dense"              # name for dense vector space


@dataclass(frozen=True)
class InferenceConfig:
    """
    Inference configuration for embedding_mode inference_* (Qdrant Inference API).

    dense_model examples:
      - "sentence-transformers/all-minilm-l6-v2"
      - "openai/text-embedding-3-large"
    sparse_model examples:
      - "Qdrant/bm25"

    dense_options / sparse_options: provider-specific options.
    Dense API key option names used by Qdrant Cloud Inference include:
      - OpenAI:     {"openai-api-key": "..."}
      - Cohere:     {"cohere-api-key": "..."}
      - Jina AI:    {"jina-api-key": "..."}
      - OpenRouter: {"openrouter-api-key": "..."}
    You can also pass provider-specific parameters like {"dimensions": 512}.
    """
    dense_model: str | None = None
    sparse_model: str | None = None
    dense_options: dict | None = None
    sparse_options: dict | None = None

    def dense_document(self, text: str) -> qcm.Document:
        if not self.dense_model:
            raise ValueError(
                "InferenceConfig.dense_model is required for inference_dense/inference_hybrid"
            )
        return qcm.Document(
            text=text,
            model=self.dense_model,
            options=self.dense_options,
        )

    def sparse_document(self, text: str) -> qcm.Document:
        if not self.sparse_model:
            raise ValueError(
                "InferenceConfig.sparse_model is required for inference_sparse/inference_hybrid"
            )
        return qcm.Document(
            text=text,
            model=self.sparse_model,
            options=self.sparse_options,
        )


class BasePomaQdrantCore:
    """
    Shared, client-agnostic core: builds payloads/points, fuses rankings, builds cheatsheet inputs.
    """

    @staticmethod
    def _to_sparse_vector(item: object) -> qmodels.SparseVector:
        if isinstance(item, qmodels.SparseVector):
            return item
        if isinstance(item, tuple) and len(item) == 2:
            indices, values = item
            return qmodels.SparseVector(indices=list(indices), values=list(values))
        if isinstance(item, dict):
            indices = item.get("indices")
            values = item.get("values")
            if indices is None or values is None:
                raise ValueError("Sparse embedding dict must contain 'indices' and 'values'")
            return qmodels.SparseVector(indices=list(indices), values=list(values))
        raise ValueError(
            "Sparse embedder must return SparseVector OR (indices, values) tuple OR dict{'indices','values'}"
        )

    @staticmethod
    def _get_file_id_fallback(chunksets: list[dict], chunks: list[dict]) -> str:
        file_id = None
        if chunksets:
            file_id = chunksets[0].get("file_id")
        if not file_id and chunks:
            file_id = chunks[0].get("file_id") or chunks[0].get("tag")
        return file_id or "unknown"

    @classmethod
    def prepare_points_from_chunk_data(
        cls,
        chunk_data: dict,
        *,
        store_chunk_details: bool = True,
    ) -> tuple[list[str], list[str], list[dict]]:
        """
        Returns: (ids, documents, payloads)
        - ids: deterministic UUID strings (file_id + chunkset_index)
        - documents: chunkset.contents
        - payloads: metadata including "text"; "chunk_details" only if store_chunk_details=True
        """
        chunksets: list[dict] = chunk_data.get("chunksets", [])
        chunks: list[dict] = chunk_data.get("chunks", [])
        if not chunksets:
            return [], [], []

        file_id = cls._get_file_id_fallback(chunksets, chunks)
        chunks_by_index = {c.get("chunk_index"): c for c in chunks}

        ids: list[str] = []
        documents: list[str] = []
        payloads: list[dict] = []

        for idx, chunkset in enumerate(chunksets):
            content = chunkset.get("contents", "")
            if not content:
                continue

            chunkset_index = chunkset.get("chunkset_index")
            if chunkset_index is None:
                chunkset_index = idx

            chunkset_file_id = chunkset.get("file_id") or file_id
            point_id = chunk_uuid_string(chunkset_file_id, int(chunkset_index))

            raw_chunks = chunkset.get("chunks", [])
            chunk_indices = list(raw_chunks) if isinstance(raw_chunks, (list, tuple)) else []

            payload: dict = {
                "chunkset_index": int(chunkset_index),
                "chunks": chunk_indices,
                "file_id": chunkset_file_id,
                "text": content,
            }
            if store_chunk_details:
                relevant_chunks = [chunks_by_index.get(ci) for ci in chunk_indices if ci in chunks_by_index]
                payload["chunk_details"] = relevant_chunks

            ids.append(point_id)
            documents.append(content)
            payloads.append(payload)

        return ids, documents, payloads

    @staticmethod
    def rrf_fuse(
        ranked_lists: list[list[str]],
        *,
        k: int,
        weights: list[float] | None = None,
    ) -> dict[str, float]:
        """
        Reciprocal Rank Fusion (optionally weighted):
          score(id) = sum_i weight_i / (k + rank_i)
        If weights is None, all weights are 1.0. Order: [dense, sparse] for hybrid.
        If weights is provided, its length must equal len(ranked_lists).
        """
        n = len(ranked_lists)
        if weights is None:
            w = [1.0] * n
        else:
            if len(weights) != n:
                raise ValueError(
                    f"rrf_fuse: len(weights)={len(weights)} must equal len(ranked_lists)={n}"
                )
            w = list(weights)
        scores: dict[str, float] = {}
        for lst, weight in zip(ranked_lists, w, strict=False):
            for rank, pid in enumerate(lst, start=1):
                scores[pid] = scores.get(pid, 0.0) + (weight / (k + rank))
        return scores

    @staticmethod
    def results_to_cheatsheet_inputs(results: list[dict]) -> tuple[list[dict], list[dict]]:
        """
        Extract relevant_chunksets and all_chunks from search results for generate_cheatsheets.

        When store_chunk_details=False, chunk_details in payload is empty, so all_chunks
        will only be populated from results. Pass chunk_data to search_cheatsheets() to
        supply full chunks for cheatsheet generation.
        """
        relevant_chunksets: list[dict] = []
        all_chunks: list[dict] = []
        seen_chunks: set[int] = set()

        for r in results:
            meta = r.get("metadata", {}) or {}
            relevant_chunksets.append(
                {
                    "chunkset_index": meta.get("chunkset_index"),
                    "chunks": meta.get("chunks", []),
                    "file_id": meta.get("file_id", "unknown"),
                }
            )

            chunk_details = meta.get("chunk_details", []) or []
            for ch in chunk_details:
                if not isinstance(ch, dict):
                    continue
                cidx = ch.get("chunk_index")
                if isinstance(cidx, int) and cidx not in seen_chunks:
                    seen_chunks.add(cidx)
                    all_chunks.append(ch)

        return relevant_chunksets, all_chunks

    @staticmethod
    def _validate_dense_size(dense_size: int | None, mode: str) -> None:
        if dense_size is None:
            raise ValueError(f'embedding_mode="{mode}" requires vectors.dense_size')
        try:
            size = int(dense_size)
        except (TypeError, ValueError) as e:
            raise ValueError(f"vectors.dense_size must be an integer, got {type(dense_size).__name__}") from e
        if size <= 0:
            raise ValueError(f'embedding_mode="{mode}" requires vectors.dense_size > 0')

    @staticmethod
    def _validate_dense_vectors(vectors: list[list[float]], expected_size: int, label: str) -> None:
        for i, v in enumerate(vectors):
            if len(v) != expected_size:
                raise ValueError(
                    f"{label} vector at index {i} has length {len(v)}, expected {expected_size} (vectors.dense_size)"
                )

    @staticmethod
    def _fastembed_dense_spec(client: object, text: str) -> tuple[str, qcm.Document]:
        dense_model = getattr(client, "embedding_model_name", None)
        if not isinstance(dense_model, str) or not dense_model:
            raise RuntimeError(
                "FastEmbed client is missing a valid embedding_model_name."
            )
        get_dense_name = getattr(client, "get_vector_field_name", None)
        if not callable(get_dense_name):
            raise RuntimeError(
                "FastEmbed client is missing get_vector_field_name()."
            )
        dense_name = get_dense_name()
        if not isinstance(dense_name, str) or not dense_name:
            raise RuntimeError(
                "FastEmbed client returned an invalid dense vector field name."
            )
        return dense_name, qcm.Document(text=text, model=dense_model)

    @staticmethod
    def _fastembed_sparse_spec(client: object, text: str) -> tuple[str, qcm.Document] | None:
        sparse_model = getattr(client, "sparse_embedding_model_name", None)
        if not sparse_model:
            return None
        get_sparse_name = getattr(client, "get_sparse_vector_field_name", None)
        if not callable(get_sparse_name):
            raise RuntimeError(
                "FastEmbed sparse model is configured but client is missing get_sparse_vector_field_name()."
            )
        sparse_name = get_sparse_name()
        if not isinstance(sparse_name, str) or not sparse_name:
            raise RuntimeError(
                "FastEmbed sparse model is configured but sparse vector field name is unavailable."
            )
        return sparse_name, qcm.Document(text=text, model=str(sparse_model))

    @classmethod
    def _fastembed_vector_payload(cls, client: object, text: str) -> dict[str, qcm.Document]:
        dense_name, dense_doc = cls._fastembed_dense_spec(client, text)
        vector: dict[str, qcm.Document] = {dense_name: dense_doc}
        sparse_spec = cls._fastembed_sparse_spec(client, text)
        if sparse_spec is not None:
            sparse_name, sparse_doc = sparse_spec
            vector[sparse_name] = sparse_doc
        return vector


class SyncClientOps:
    """
    Minimal sync wrappers for the QdrantClient I/O surface used by this module.
    """

    def __init__(self, client: QdrantClient):
        self.client = client

    def get_collections_names(self) -> set[str]:
        return {c.name for c in self.client.get_collections().collections}

    def create_collection_dense(self, collection_name: str, *, vectors: VectorConfig) -> None:
        self.client.create_collection(
            collection_name=collection_name,
            vectors_config={
                vectors.dense_name: qmodels.VectorParams(size=int(vectors.dense_size), distance=vectors.distance)
            },
        )

    def create_collection_sparse(self, collection_name: str, *, vectors: VectorConfig) -> None:
        try:
            self.client.create_collection(
                collection_name=collection_name,
                vectors_config={},
                sparse_vectors_config={vectors.sparse_name: qmodels.SparseVectorParams()},
            )
        except Exception as e:
            raise RuntimeError(
                "Creating a sparse-only collection failed (some Qdrant setups require at least one dense vector). "
                "Use embedding_mode='external_hybrid' with dense + sparse if needed."
            ) from e

    def create_collection_hybrid(self, collection_name: str, *, vectors: VectorConfig) -> None:
        self.client.create_collection(
            collection_name=collection_name,
            vectors_config={
                vectors.dense_name: qmodels.VectorParams(size=int(vectors.dense_size), distance=vectors.distance)
            },
            sparse_vectors_config={vectors.sparse_name: qmodels.SparseVectorParams()},
        )

    def ensure_fastembed_collection(self, collection_name: str) -> None:
        existing = self.get_collections_names()
        if collection_name in existing:
            return

        get_dense_params = getattr(self.client, "get_fastembed_vector_params", None)
        if not callable(get_dense_params):
            raise RuntimeError(
                "FastEmbed client is missing get_fastembed_vector_params()."
            )
        vectors_config = get_dense_params()
        if not vectors_config:
            raise RuntimeError(
                "FastEmbed client returned empty dense vector params."
            )

        get_sparse_params = getattr(self.client, "get_fastembed_sparse_vector_params", None)
        sparse_vectors_config = (
            get_sparse_params() if callable(get_sparse_params) else None
        )

        kwargs: dict = {
            "collection_name": collection_name,
            "vectors_config": vectors_config,
        }
        if sparse_vectors_config:
            kwargs["sparse_vectors_config"] = sparse_vectors_config
        self.client.create_collection(**kwargs)

    def add_documents(self, collection_name: str, *, documents: list[str], payloads: list[dict], ids: list[str]) -> None:
        self.ensure_fastembed_collection(collection_name)
        points: list[qmodels.PointStruct] = []
        for pid, doc, payload in zip(ids, documents, payloads, strict=True):
            vector = BasePomaQdrantCore._fastembed_vector_payload(self.client, doc)
            points.append(qmodels.PointStruct(id=pid, vector=vector, payload=payload))
        self.upsert_points(collection_name, points=points)

    def upsert_points(self, collection_name: str, *, points: list[qmodels.PointStruct]) -> None:
        self.client.upsert(collection_name=collection_name, points=points)

    def retrieve_ids(self, collection_name: str, *, ids: list[str]) -> set[str]:
        """Return the set of point ids that exist in the collection (for insert_only write_mode)."""
        if not ids:
            return set()
        try:
            points = self.client.retrieve(collection_name=collection_name, ids=ids)
            return {str(p.id) for p in points}
        except Exception:
            return set()

    def query_text(
        self,
        collection_name: str,
        *,
        query: str,
        limit: int,
        query_filter: qmodels.Filter | None = None,
    ) -> list[dict]:
        dense_name, dense_doc = BasePomaQdrantCore._fastembed_dense_spec(
            self.client, query
        )
        return self.query_points(
            collection_name,
            query=dense_doc,
            using=dense_name,
            limit=limit,
            query_filter=query_filter,
        )

    def search_named_vector(
        self,
        collection_name: str,
        *,
        name: str,
        vector: object,
        limit: int,
        query_filter: qmodels.Filter | None = None,
    ) -> list[dict]:
        return self.query_points(
            collection_name,
            query=vector,
            using=name,
            limit=limit,
            query_filter=query_filter,
        )

    def query_points(
        self,
        collection_name: str,
        *,
        query: object,
        using: str | None,
        limit: int,
        query_filter: qmodels.Filter | None = None,
    ) -> list[dict]:
        kwargs: dict = {
            "collection_name": collection_name,
            "query": query,
            "limit": limit,
            "with_payload": True,
        }
        if using is not None:
            kwargs["using"] = using
        try:
            res = self.client.query_points(**kwargs, query_filter=query_filter)
        except TypeError:
            res = self.client.query_points(**kwargs, filter=query_filter)
        hits = getattr(res, "points", res) if not isinstance(res, list) else res
        out: list[dict] = []
        for h in hits:
            payload = getattr(h, "payload", None) or {}
            out.append(
                {
                    "id": str(getattr(h, "id", "") or ""),
                    "score": float(getattr(h, "score", 0.0) or 0.0),
                    "text": payload.get("text", "") if isinstance(payload, dict) else "",
                    "metadata": payload if isinstance(payload, dict) else {},
                }
            )
        return out

    def query_points_inference(
        self,
        collection_name: str,
        *,
        query: object,
        using: str | None,
        limit: int,
        query_filter: qmodels.Filter | None = None,
    ) -> list[dict]:
        """Query using an inference object (e.g. qcm.Document)."""
        return self.query_points(
            collection_name,
            query=query,
            using=using,
            limit=limit,
            query_filter=query_filter,
        )


class PomaQdrant(BasePomaQdrantCore):
    """
    Sync implementation.

    embedding_mode:
      - "fastembed"        : upsert/query_points with models.Document
      - "external_dense"   : dense_embed_fn + upsert/search on named dense vector
      - "external_sparse"  : sparse_embed_fn + upsert/search on named sparse vector
      - "external_hybrid"  : both embedders + two searches + RRF fusion
      - "inference_dense"   : Qdrant Inference API (inference_config.dense_model)
      - "inference_sparse" : Qdrant Inference API (inference_config.sparse_model)
      - "inference_hybrid" : Qdrant Inference API (dense + sparse + RRF)

    fastembed: Requires a client with FastEmbed integration (upsert/query_points with models.Document).
    Not guaranteed in all environments; for production, pass a
    pre-configured qdrant_client or use external_* mode. See _ensure_fastembed_client().

    write_mode: "upsert" (default) or "insert_only" (skip points that already exist).
    batch_size: chunk ingest into batches of this size (default 256).
    store_chunk_details: if True (default), payload includes chunk_details so search_cheatsheets can build cheatsheets without chunk_data.
    """

    _INFERENCE_MODES = {"inference_dense", "inference_sparse", "inference_hybrid"}
    _DENSE_PROVIDER_KEY_RULES: tuple[
        tuple[tuple[str, ...], str, tuple[str, ...], str],
        ...,
    ] = (
        (("openai/",), "openai-api-key", ("OPENAI_API_KEY",), "OpenAI"),
        (("cohere/",), "cohere-api-key", ("COHERE_API_KEY",), "Cohere"),
        (("jinaai/", "jina/"), "jina-api-key", ("JINA_API_KEY", "JINAAI_API_KEY"), "Jina AI"),
        (("openrouter/",), "openrouter-api-key", ("OPENROUTER_API_KEY",), "OpenRouter"),
    )

    @staticmethod
    def _resolve_cloud_inference_enabled(
        qdrant_client: QdrantClient | None,
        cfg: QdrantConfig,
    ) -> bool:
        if qdrant_client is None:
            return bool(cfg.cloud_inference)
        client_flag = getattr(qdrant_client, "cloud_inference", None)
        if client_flag is None:
            return bool(cfg.cloud_inference)
        return bool(client_flag)

    @staticmethod
    def _copy_options_dict(options: dict | None, field_name: str) -> dict:
        if options is None:
            return {}
        if not isinstance(options, dict):
            raise ValueError(
                f"InferenceConfig.{field_name} must be a dict, got {type(options).__name__}"
            )
        return dict(options)

    @staticmethod
    def _with_dense_size(vectors: VectorConfig, dense_size: int) -> VectorConfig:
        return VectorConfig(
            distance=vectors.distance,
            dense_size=dense_size,
            sparse_name=vectors.sparse_name,
            dense_name=vectors.dense_name,
        )

    @staticmethod
    def _copy_client_kwargs(client_kwargs: dict[str, object] | None) -> dict[str, object]:
        if client_kwargs is None:
            return {}
        if not isinstance(client_kwargs, dict):
            raise ValueError(
                f"qdrant.client_kwargs must be a dict, got {type(client_kwargs).__name__}"
            )
        return dict(client_kwargs)

    @staticmethod
    def _validate_client_kwargs_conflicts(mode: str, client_kwargs: dict[str, object]) -> None:
        blocked_by_mode = {
            "url": {"url", "api_key", "cloud_inference"},
            "path": {"path"},
            "memory": {"location"},
        }
        blocked = blocked_by_mode.get(mode, set())
        conflicts = sorted(k for k in client_kwargs if k in blocked)
        if not conflicts:
            return
        joined = ", ".join(conflicts)
        raise ValueError(
            f'qdrant.client_kwargs for mode="{mode}" cannot include: {joined}. '
            "Set these values via QdrantConfig fields instead."
        )

    def _apply_dense_provider_api_key_defaults(
        self,
        *,
        dense_model: str,
        dense_options: dict,
    ) -> None:
        model_name = dense_model.strip().lower()
        for prefixes, option_key, env_names, provider_name in self._DENSE_PROVIDER_KEY_RULES:
            if not any(model_name.startswith(prefix) for prefix in prefixes):
                continue
            # Explicit user config always wins.
            if option_key in dense_options:
                return
            for env_name in env_names:
                env_api_key = os.getenv(env_name)
                if env_api_key:
                    dense_options[option_key] = env_api_key
                    return
            env_hint = " or ".join(env_names)
            raise ValueError(
                f"cloud_inference=True with {provider_name} dense inference requires an API key. "
                f"Set InferenceConfig.dense_options['{option_key}'] or {env_hint}."
            )
        # Unknown provider prefix: leave unchanged.

    def _apply_cloud_inference_defaults(self) -> None:
        dense_mode = self.embedding_mode in {"inference_dense", "inference_hybrid"}
        sparse_mode = self.embedding_mode in {"inference_sparse", "inference_hybrid"}

        cfg = self.inference_config or InferenceConfig()
        dense_model = cfg.dense_model
        sparse_model = cfg.sparse_model
        dense_options = self._copy_options_dict(cfg.dense_options, "dense_options")
        sparse_options = self._copy_options_dict(cfg.sparse_options, "sparse_options")

        if dense_mode and not dense_model:
            dense_model = DEFAULT_CLOUD_DENSE_MODEL

        if sparse_mode and not sparse_model:
            sparse_model = DEFAULT_CLOUD_SPARSE_MODEL

        normalized_vector_dense_size: int | None = None
        if dense_mode and self.vectors.dense_size is not None:
            try:
                normalized_vector_dense_size = int(self.vectors.dense_size)
            except (TypeError, ValueError) as e:
                raise ValueError("vectors.dense_size must be an integer") from e
            if normalized_vector_dense_size <= 0:
                raise ValueError("vectors.dense_size must be > 0")
            if normalized_vector_dense_size != self.vectors.dense_size:
                self.vectors = self._with_dense_size(
                    self.vectors, normalized_vector_dense_size
                )

        normalized_dense_dim: int | None = None
        if dense_mode and dense_model:
            dense_dim_raw = dense_options.get("dimensions")
            if dense_dim_raw is not None:
                try:
                    normalized_dense_dim = int(dense_dim_raw)
                except (TypeError, ValueError) as e:
                    raise ValueError(
                        "InferenceConfig.dense_options['dimensions'] must be an integer"
                    ) from e
                if normalized_dense_dim <= 0:
                    raise ValueError(
                        "InferenceConfig.dense_options['dimensions'] must be > 0"
                    )
                dense_options["dimensions"] = normalized_dense_dim

            # Allow setting dimension in exactly one place (vectors OR dense_options).
            if (
                dense_model == DEFAULT_CLOUD_DENSE_MODEL
                and normalized_dense_dim is None
                and normalized_vector_dense_size is None
            ):
                normalized_dense_dim = DEFAULT_CLOUD_DENSE_DIMENSIONS
                dense_options["dimensions"] = normalized_dense_dim

            if normalized_dense_dim is None and normalized_vector_dense_size is not None:
                normalized_dense_dim = normalized_vector_dense_size
                if dense_model.startswith("openai/"):
                    dense_options["dimensions"] = normalized_dense_dim
            elif normalized_dense_dim is not None and normalized_vector_dense_size is None:
                self.vectors = self._with_dense_size(
                    self.vectors, normalized_dense_dim
                )
            elif (
                normalized_dense_dim is not None
                and normalized_vector_dense_size is not None
                and normalized_dense_dim != normalized_vector_dense_size
            ):
                raise ValueError(
                    "vectors.dense_size must match InferenceConfig.dense_options['dimensions']: "
                    f"{normalized_vector_dense_size} != {normalized_dense_dim}"
                )

            self._apply_dense_provider_api_key_defaults(
                dense_model=dense_model,
                dense_options=dense_options,
            )

        self.inference_config = InferenceConfig(
            dense_model=dense_model,
            sparse_model=sparse_model,
            dense_options=dense_options or None,
            sparse_options=sparse_options or None,
        )

    def _raise_qdrant_response(
        self,
        exc: UnexpectedResponse,
        *,
        operation: str,
    ) -> None:
        content = exc.content.decode("utf-8", errors="replace")
        message = f"Qdrant response error during `{operation}`: {exc}"
        lowered = f"{exc}\n{content}".lower()
        if self.cloud_inference and (
            "endpoint configuration not found for model" in lowered
            or "inference can't process request" in lowered
        ):
            message = (
                f"{message}. The model/provider is not configured for this Qdrant Cloud cluster "
                "(check the cluster Inference tab)."
            )
        raise QdrantResponseError(
            message,
            status_code=exc.status_code,
            reason_phrase=exc.reason_phrase,
            raw_content=content,
        ) from exc

    def __init__(
        self,
        collection_name: str = "poma_documents",
        *,
        qdrant: QdrantConfig | None = None,
        qdrant_client: QdrantClient | None = None,
        embedding_mode: str | None = None,
        dense_embed_fn: DenseEmbedSync | None = None,
        sparse_embed_fn: SparseEmbedSync | None = None,
        vectors: VectorConfig | None = None,
        inference_config: InferenceConfig | None = None,
        fusion_k: int = 60,
        fusion_candidates: int = 50,
        fusion_weights: tuple[float, float] | None = None,
        write_mode: str = "upsert",
        batch_size: int = 256,
        store_chunk_details: bool = True,
    ):
        self.collection_name = collection_name
        cfg = qdrant or QdrantConfig()
        self.cloud_inference = self._resolve_cloud_inference_enabled(qdrant_client, cfg)
        mode_text = (embedding_mode or "").strip().lower()
        if mode_text:
            self.embedding_mode = mode_text
        else:
            # Strong default: cloud inference uses inference_dense; otherwise keep fastembed.
            self.embedding_mode = "inference_dense" if self.cloud_inference else "fastembed"
        self.dense_embed_fn = dense_embed_fn
        self.sparse_embed_fn = sparse_embed_fn
        self.vectors = vectors or VectorConfig()
        self.inference_config = inference_config
        self.fusion_k = max(1, int(fusion_k))
        self.fusion_candidates = max(1, int(fusion_candidates))
        self.fusion_weights = fusion_weights  # (dense_weight, sparse_weight) for hybrid; None = (1, 1)
        self.write_mode = (write_mode or "upsert").strip().lower()
        self.batch_size = max(1, int(batch_size))
        self.store_chunk_details = bool(store_chunk_details)
        if self.write_mode not in ("upsert", "insert_only"):
            raise ValueError('write_mode must be one of: "upsert", "insert_only"')

        allowed = {
            "fastembed", "external_dense", "external_sparse", "external_hybrid",
            "inference_dense", "inference_sparse", "inference_hybrid",
        }
        if self.embedding_mode not in allowed:
            raise ValueError(f"embedding_mode must be one of: {sorted(allowed)}")

        if self.cloud_inference and self.embedding_mode == "fastembed":
            raise ValueError(
                "cloud_inference=True is not supported with embedding_mode='fastembed'. "
                "Use cloud_inference=False for local FastEmbed or switch to an inference_* mode "
                "with InferenceConfig."
            )
        if self.cloud_inference and self.embedding_mode in self._INFERENCE_MODES:
            self._apply_cloud_inference_defaults()

        # Fail-fast: validate external_* requirements at init
        if self.embedding_mode == "external_dense":
            if self.dense_embed_fn is None:
                raise ValueError('embedding_mode="external_dense" requires dense_embed_fn')
            self._validate_dense_size(self.vectors.dense_size, "external_dense")
        elif self.embedding_mode == "external_sparse":
            if self.sparse_embed_fn is None:
                raise ValueError('embedding_mode="external_sparse" requires sparse_embed_fn')
        elif self.embedding_mode == "external_hybrid":
            if self.dense_embed_fn is None or self.sparse_embed_fn is None:
                raise ValueError('embedding_mode="external_hybrid" requires dense_embed_fn AND sparse_embed_fn')
            self._validate_dense_size(self.vectors.dense_size, "external_hybrid")
        elif self.embedding_mode == "inference_dense":
            if self.inference_config is None:
                raise ValueError('embedding_mode="inference_dense" requires inference_config')
            self.inference_config.dense_document("")  # validate dense_model set
            self._validate_dense_size(self.vectors.dense_size, "inference_dense")
        elif self.embedding_mode == "inference_sparse":
            if self.inference_config is None:
                raise ValueError('embedding_mode="inference_sparse" requires inference_config')
            self.inference_config.sparse_document("")  # validate sparse_model set
        elif self.embedding_mode == "inference_hybrid":
            if self.inference_config is None:
                raise ValueError('embedding_mode="inference_hybrid" requires inference_config')
            self.inference_config.dense_document("")
            self.inference_config.sparse_document("")
            self._validate_dense_size(self.vectors.dense_size, "inference_hybrid")

        if qdrant_client is not None:
            self.client = qdrant_client
        else:
            mode = (cfg.mode or "").strip().lower()
            client_kwargs = self._copy_client_kwargs(cfg.client_kwargs)
            if mode == "url":
                self._validate_client_kwargs_conflicts(mode, client_kwargs)
                if not cfg.url:
                    raise ValueError('qdrant.mode="url" requires qdrant.url="http(s)://..."')
                self.client = QdrantClient(
                    url=cfg.url,
                    api_key=cfg.api_key,
                    cloud_inference=cfg.cloud_inference,
                    **client_kwargs,
                )
            elif mode == "path":
                self._validate_client_kwargs_conflicts(mode, client_kwargs)
                if not cfg.path:
                    raise ValueError('qdrant.mode="path" requires qdrant.path="..."')
                self.client = QdrantClient(path=cfg.path, **client_kwargs)
            elif mode == "memory":
                self._validate_client_kwargs_conflicts(mode, client_kwargs)
                self.client = QdrantClient(location=":memory:", **client_kwargs)
            else:
                raise ValueError('qdrant.mode must be one of: "memory", "path", "url"')

        self.ops = SyncClientOps(self.client)
        self._ensure_collection_if_needed_sync()

    def _ensure_fastembed_client(self) -> None:
        """Raise with a clear message if fastembed mode is used but client lacks FastEmbed Document support."""
        query_points_fn = getattr(self.client, "query_points", None)
        upsert_fn = getattr(self.client, "upsert", None)
        get_dense_name_fn = getattr(self.client, "get_vector_field_name", None)
        get_dense_params_fn = getattr(self.client, "get_fastembed_vector_params", None)
        if (
            not callable(query_points_fn)
            or not callable(upsert_fn)
            or not callable(get_dense_name_fn)
            or not callable(get_dense_params_fn)
        ):
            raise RuntimeError(
                "embedding_mode='fastembed' requires a Qdrant client with FastEmbed integration "
                "(client.query_points(..., query=models.Document(...)) and client.upsert(...)). "
                "Pass a pre-configured client or use embedding_mode='external_dense' (or external_sparse/external_hybrid) "
                "with your own embedder."
            )
        dense_model = getattr(self.client, "embedding_model_name", None)
        if not isinstance(dense_model, str) or not dense_model:
            raise RuntimeError(
                "embedding_mode='fastembed' requires a valid embedding_model_name on the client."
            )
        try:
            sig_query_points = inspect.signature(query_points_fn)
            sig_upsert = inspect.signature(upsert_fn)
        except (ValueError, TypeError):
            raise RuntimeError(
                "embedding_mode='fastembed': client query_points/upsert could not be inspected. "
                "Use a Qdrant client with FastEmbed integration or embedding_mode='external_dense'."
            )
        if "query" not in sig_query_points.parameters:
            raise RuntimeError(
                "embedding_mode='fastembed' requires client.query_points(..., query=..., ...). "
                "This client's query_points() does not accept 'query'. Use a FastEmbed-enabled client or external_* mode."
            )
        if "points" not in sig_upsert.parameters:
            raise RuntimeError(
                "embedding_mode='fastembed' requires client.upsert(..., points=...). "
                "This client's upsert() does not accept 'points'. Use a FastEmbed-enabled client or external_* mode."
            )

    def _ensure_collection_if_needed_sync(self) -> None:
        existing = self.ops.get_collections_names()
        if self.collection_name in existing:
            return

        if self.embedding_mode in ("external_dense", "inference_dense"):
            self.ops.create_collection_dense(self.collection_name, vectors=self.vectors)
        elif self.embedding_mode in ("external_sparse", "inference_sparse"):
            self.ops.create_collection_sparse(self.collection_name, vectors=self.vectors)
        elif self.embedding_mode in ("external_hybrid", "inference_hybrid"):
            self.ops.create_collection_hybrid(self.collection_name, vectors=self.vectors)
        # fastembed: created lazily in SyncClientOps.add_documents().

    def _filter_new_ids_sync(self, ids: list[str]) -> tuple[list[str], list[int]]:
        """For insert_only: return (ids that are new, indices into original list)."""
        existing = self.ops.retrieve_ids(self.collection_name, ids=ids)
        new_ids = [pid for pid in ids if pid not in existing]
        new_set = set(new_ids)
        indices = [i for i, pid in enumerate(ids) if pid in new_set]
        return new_ids, indices

    def ingest(
        self,
        chunk_data: dict | str | os.PathLike[str],
    ) -> None:
        """
        Ingest chunk data into the collection.
        chunk_data: either a dict with ``chunks`` and ``chunksets`` keys,
        or a path to a ``.poma`` file (will be extracted first).
        """
        try:
            chunk_data = _chunk_data_from_input(chunk_data)
            ids, documents, payloads = self.prepare_points_from_chunk_data(
                chunk_data, store_chunk_details=self.store_chunk_details
            )
            if not documents:
                return

            if self.embedding_mode != "fastembed":
                self._ensure_collection_if_needed_sync()

            if self.write_mode == "insert_only" and ids:
                new_ids, indices = self._filter_new_ids_sync(ids)
                if not new_ids:
                    return
                ids = new_ids
                documents = [documents[i] for i in indices]
                payloads = [payloads[i] for i in indices]

            if self.embedding_mode == "fastembed":
                self._ensure_fastembed_client()
                for start in range(0, len(ids), self.batch_size):
                    end = start + self.batch_size
                    self.ops.add_documents(
                        self.collection_name,
                        documents=documents[start:end],
                        payloads=payloads[start:end],
                        ids=ids[start:end],
                    )
                return

            if self.embedding_mode == "external_dense":
                dense_vecs = self.dense_embed_fn(documents)
                if len(dense_vecs) != len(documents):
                    raise ValueError("dense_embed_fn returned a different number of vectors than documents")
                self._validate_dense_vectors(dense_vecs, int(self.vectors.dense_size), "dense_embed_fn")
                points = [
                    qmodels.PointStruct(id=pid, vector={self.vectors.dense_name: vec}, payload=payload)
                    for pid, vec, payload in zip(ids, dense_vecs, payloads, strict=True)
                ]
                for start in range(0, len(points), self.batch_size):
                    self.ops.upsert_points(self.collection_name, points=points[start : start + self.batch_size])
                return

            if self.embedding_mode == "external_sparse":
                sparse_items = self.sparse_embed_fn(documents)
                if len(sparse_items) != len(documents):
                    raise ValueError("sparse_embed_fn returned a different number of vectors than documents")
                points: list[qmodels.PointStruct] = []
                for pid, item, payload in zip(ids, sparse_items, payloads, strict=True):
                    svec = self._to_sparse_vector(item)
                    points.append(
                        qmodels.PointStruct(id=pid, vector={self.vectors.sparse_name: svec}, payload=payload)
                    )
                for start in range(0, len(points), self.batch_size):
                    self.ops.upsert_points(self.collection_name, points=points[start : start + self.batch_size])
                return

            if self.embedding_mode == "inference_dense":
                points_inf = [
                    qmodels.PointStruct(
                        id=pid,
                        vector={self.vectors.dense_name: self.inference_config.dense_document(doc)},
                        payload=payload,
                    )
                    for pid, doc, payload in zip(ids, documents, payloads, strict=True)
                ]
                for start in range(0, len(points_inf), self.batch_size):
                    self.ops.upsert_points(self.collection_name, points=points_inf[start : start + self.batch_size])
                return

            if self.embedding_mode == "inference_sparse":
                points_inf = [
                    qmodels.PointStruct(
                        id=pid,
                        vector={self.vectors.sparse_name: self.inference_config.sparse_document(doc)},
                        payload=payload,
                    )
                    for pid, doc, payload in zip(ids, documents, payloads, strict=True)
                ]
                for start in range(0, len(points_inf), self.batch_size):
                    self.ops.upsert_points(self.collection_name, points=points_inf[start : start + self.batch_size])
                return

            if self.embedding_mode == "inference_hybrid":
                points_inf = [
                    qmodels.PointStruct(
                        id=pid,
                        vector={
                            self.vectors.dense_name: self.inference_config.dense_document(doc),
                            self.vectors.sparse_name: self.inference_config.sparse_document(doc),
                        },
                        payload=payload,
                    )
                    for pid, doc, payload in zip(ids, documents, payloads, strict=True)
                ]
                for start in range(0, len(points_inf), self.batch_size):
                    self.ops.upsert_points(self.collection_name, points=points_inf[start : start + self.batch_size])
                return

            # external_hybrid
            dense_vecs = self.dense_embed_fn(documents)
            sparse_items = self.sparse_embed_fn(documents)
            if len(dense_vecs) != len(documents) or len(sparse_items) != len(documents):
                raise ValueError("Hybrid embedders returned a different number of vectors than documents")
            self._validate_dense_vectors(dense_vecs, int(self.vectors.dense_size), "dense_embed_fn")
            points_list: list[qmodels.PointStruct] = []
            for pid, dvec, sitem, payload in zip(ids, dense_vecs, sparse_items, payloads, strict=True):
                svec = self._to_sparse_vector(sitem)
                points_list.append(
                    qmodels.PointStruct(
                        id=pid,
                        vector={self.vectors.dense_name: dvec, self.vectors.sparse_name: svec},
                        payload=payload,
                    )
                )
            for start in range(0, len(points_list), self.batch_size):
                self.ops.upsert_points(self.collection_name, points=points_list[start : start + self.batch_size])
        except UnexpectedResponse as exc:
            self._raise_qdrant_response(exc, operation="ingest")

    def search(
        self,
        query: str,
        *,
        limit: int = 5,
        query_filter: qmodels.Filter | None = None,
    ) -> list[dict]:
        """Search; returns list[dict] with keys id, score, text, metadata (same shape as SearchResult)."""
        limit = max(1, int(limit))
        try:
            if self.embedding_mode == "fastembed":
                self._ensure_fastembed_client()
                dense_name, dense_doc = self._fastembed_dense_spec(self.client, query)
                sparse_spec = self._fastembed_sparse_spec(self.client, query)
                if sparse_spec is None:
                    return self.ops.query_points(
                        self.collection_name,
                        query=dense_doc,
                        using=dense_name,
                        limit=limit,
                        query_filter=query_filter,
                    )

                cand = max(limit, self.fusion_candidates)
                sparse_name, sparse_doc = sparse_spec
                dense_hits = self.ops.query_points(
                    self.collection_name,
                    query=dense_doc,
                    using=dense_name,
                    limit=cand,
                    query_filter=query_filter,
                )
                sparse_hits = self.ops.query_points(
                    self.collection_name,
                    query=sparse_doc,
                    using=sparse_name,
                    limit=cand,
                    query_filter=query_filter,
                )
                dense_ids = [str(h["id"]) for h in dense_hits]
                sparse_ids = [str(h["id"]) for h in sparse_hits]
                fused = self.rrf_fuse([dense_ids, sparse_ids], k=self.fusion_k, weights=None)
                payload_by_id: dict[str, dict] = {}
                for h in dense_hits + sparse_hits:
                    payload_by_id.setdefault(str(h["id"]), h.get("metadata", {}) or {})
                ranked = sorted(fused.items(), key=lambda kv: kv[1], reverse=True)[:limit]
                return [
                    {
                        "id": pid,
                        "score": score,
                        "text": payload_by_id.get(pid, {}).get("text", ""),
                        "metadata": payload_by_id.get(pid, {}) or {},
                    }
                    for pid, score in ranked
                ]

            self._ensure_collection_if_needed_sync()

            if self.embedding_mode == "external_dense":
                qvec = self.dense_embed_fn([query])[0]
                return self.ops.search_named_vector(
                    self.collection_name,
                    name=self.vectors.dense_name,
                    vector=qvec,
                    limit=limit,
                    query_filter=query_filter,
                )

            if self.embedding_mode == "external_sparse":
                sitem = self.sparse_embed_fn([query])[0]
                svec = self._to_sparse_vector(sitem)
                return self.ops.search_named_vector(
                    self.collection_name,
                    name=self.vectors.sparse_name,
                    vector=svec,
                    limit=limit,
                    query_filter=query_filter,
                )

            if self.embedding_mode == "inference_dense":
                query_doc = self.inference_config.dense_document(query)
                return self.ops.query_points_inference(
                    self.collection_name,
                    query=query_doc,
                    using=self.vectors.dense_name,
                    limit=limit,
                    query_filter=query_filter,
                )

            if self.embedding_mode == "inference_sparse":
                query_doc = self.inference_config.sparse_document(query)
                return self.ops.query_points_inference(
                    self.collection_name,
                    query=query_doc,
                    using=self.vectors.sparse_name,
                    limit=limit,
                    query_filter=query_filter,
                )

            if self.embedding_mode == "inference_hybrid":
                cand = max(limit, self.fusion_candidates)
                weights = list(self.fusion_weights) if self.fusion_weights else None
                dense_doc = self.inference_config.dense_document(query)
                sparse_doc = self.inference_config.sparse_document(query)
                dense_hits = self.ops.query_points_inference(
                    self.collection_name,
                    query=dense_doc,
                    using=self.vectors.dense_name,
                    limit=cand,
                    query_filter=query_filter,
                )
                sparse_hits = self.ops.query_points_inference(
                    self.collection_name,
                    query=sparse_doc,
                    using=self.vectors.sparse_name,
                    limit=cand,
                    query_filter=query_filter,
                )
                dense_ids = [str(h["id"]) for h in dense_hits]
                sparse_ids = [str(h["id"]) for h in sparse_hits]
                fused = self.rrf_fuse([dense_ids, sparse_ids], k=self.fusion_k, weights=weights)
                payload_by_id: dict[str, dict] = {}
                for h in dense_hits + sparse_hits:
                    payload_by_id.setdefault(str(h["id"]), h.get("metadata", {}) or {})
                ranked = sorted(fused.items(), key=lambda kv: kv[1], reverse=True)[:limit]
                return [
                    {
                        "id": pid,
                        "score": score,
                        "text": payload_by_id.get(pid, {}).get("text", ""),
                        "metadata": payload_by_id.get(pid, {}) or {},
                    }
                    for pid, score in ranked
                ]

            # external_hybrid: two searches + weighted RRF fusion
            cand = max(limit, self.fusion_candidates)
            weights = list(self.fusion_weights) if self.fusion_weights else None

            dense_q = self.dense_embed_fn([query])[0]
            sparse_item = self.sparse_embed_fn([query])[0]
            sparse_q = self._to_sparse_vector(sparse_item)

            dense_hits = self.ops.search_named_vector(
                self.collection_name,
                name=self.vectors.dense_name,
                vector=dense_q,
                limit=cand,
                query_filter=query_filter,
            )
            sparse_hits = self.ops.search_named_vector(
                self.collection_name,
                name=self.vectors.sparse_name,
                vector=sparse_q,
                limit=cand,
                query_filter=query_filter,
            )

            dense_ids = [str(h["id"]) for h in dense_hits]
            sparse_ids = [str(h["id"]) for h in sparse_hits]
            fused = self.rrf_fuse([dense_ids, sparse_ids], k=self.fusion_k, weights=weights)

            payload_by_id: dict[str, dict] = {}
            for h in dense_hits + sparse_hits:
                payload_by_id.setdefault(str(h["id"]), h.get("metadata", {}) or {})

            ranked = sorted(fused.items(), key=lambda kv: kv[1], reverse=True)[:limit]
            return [
                {
                    "id": pid,
                    "score": score,
                    "text": payload_by_id.get(pid, {}).get("text", ""),
                    "metadata": payload_by_id.get(pid, {}) or {},
                }
                for pid, score in ranked
            ]
        except UnexpectedResponse as exc:
            self._raise_qdrant_response(exc, operation="search")

    def close(self) -> None:
        """Close the Qdrant client connection (e.g. HTTP session). No-op if client has no close."""
        if hasattr(self.client, "close") and callable(self.client.close):
            self.client.close()

    def search_cheatsheets(
        self,
        query: str,
        *,
        limit: int = 5,
        chunk_data: dict | None = None,
        query_filter: qmodels.Filter | None = None,
    ) -> list[dict]:
        results = self.search(query, limit=limit, query_filter=query_filter)
        if not results:
            return []
        relevant_chunksets, all_chunks = self.results_to_cheatsheet_inputs(results)
        if chunk_data:
            all_chunks = chunk_data.get("chunks", []) or []
        return generate_cheatsheets(relevant_chunksets, all_chunks) or []

    # Legacy compatibility
    def store_chunksets(self, chunksets: list[dict], chunks: list[dict], *, file_id: str | None = None) -> None:
        self.ingest({"chunksets": chunksets, "chunks": chunks})

    def write(self, chunk_data: dict, *, file_id: str | None = None) -> None:
        self.ingest(chunk_data)

    def query_chunksets(self, query: str, *, top_k: int = 5) -> list[dict]:
        return [r["metadata"] for r in self.search(query, limit=top_k)]

    def query_and_generate_cheatsheet(self, query: str, *, top_k: int = 5, all_chunks: list[dict] | None = None):
        chunk_data = {"chunks": all_chunks} if all_chunks else None
        return self.search_cheatsheets(query, limit=top_k, chunk_data=chunk_data)
